"""
    请求记录中间件

    目前记录：
        请求数
        请求成功数
        请求失败数

    注意：双继承意味着双导入，所以不能在 init 上共享信息
"""
import json
from typing import Union
from palp import settings
from palp.network.request import Request
from palp.decorator.decorator_lock import FuncLock
from palp.middleware.middleware_spider import SpiderMiddleware
from palp.middleware.middleware_request import RequestMiddleware


class RequestsRecordMiddleware(RequestMiddleware, SpiderMiddleware):
    """
        请求记录中间件
    """
    request_count_all = 0  # 请求总数
    request_count_failed = 0  # 失败总数
    request_count_succeed = 0  # 成功总数

    @FuncLock()
    def request_in(self, spider, request) -> None:
        """
            记录进入的请求
        """
        self.__class__.request_count_all += 1

    @FuncLock()
    def request_failed(self, spider, request) -> None:
        """
            记录失败的请求
        """
        self.__class__.request_count_failed += 1

    @FuncLock()
    def request_close(self, spider, request, response) -> Union[Request, None]:
        """
            记录成功的请求
        """
        self.__class__.request_count_succeed += 1

        return

    def spider_close(self, spider) -> None:
        """
            直接双继承，通过 spider_close 上传记录信息
        """
        from palp.conn import redis_conn

        if redis_conn is not None:
            stop_time = redis_conn.get(settings.REDIS_KEY_STOP)
            redis_conn.set(settings.REDIS_KEY_STOP, json.dumps({
                'request_all': self.__class__.request_count_all,
                'request_failed': self.__class__.request_count_failed,
                'request_succeed': self.__class__.request_count_succeed,
                'stop_time': stop_time.decode()
            }, ensure_ascii=False))



"""
    request 失败回收中间件
"""
import json
from palp import settings
from palp.middleware.middleware_request import RequestMiddleware


class RequestRecycleMiddleware(RequestMiddleware):
    """
        失败请求回收中间件
    """

    def request_failed(self, spider, request):
        """
        回收失败的请求，为了能够直观的展示，就不使用 pickle 了

        注意：开启 REQUEST_FAILED_SAVE 才会启用

        :param spider:
        :param request:
        :return:
        """
        from palp.conn import redis_conn

        if redis_conn is None:
            return

        # 保存请求
        req = request.to_dict()

        # 转化 callback
        if request.callback:
            req['callback'] = req['callback'].__name__

        # 转化 cookie_jar
        if request.cookie_jar:
            if request.cookies:
                request.cookie_jar.update(request.cookies)
            req['cookies'] = request.cookie_jar.get_dict()
            del req['cookie_jar']

        # 转化 downloader
        if request.downloader != request.__class__.DOWNLOADER:
            req['downloader'] = {
                'module': request.downloader.__module__,  # 引用自哪里（downloader 都是未实例化的所以不需要 __class__）
                'init': request.downloader.__name__,  # 模块名叫什么
            }
        else:
            del req['downloader']

        # 转化 downloader_parser
        if request.downloader_parser != request.__class__.DOWNLOADER_PARSER:
            req['downloader_parser'] = {
                'module': request.downloader_parser.__module__,  # 引用自哪里
                'init': request.downloader_parser.__name__,  # 模块名叫什么
            }
        else:
            del req['downloader_parser']

        # 删除代理
        if request.proxies:
            del req['proxies']

        redis_conn.sadd(settings.REDIS_KEY_QUEUE_BAD_REQUEST, json.dumps(req, ensure_ascii=False))



"""
    严格的 item（类似 scrapy item）

    示例
        class ItemStrict(palp.ItemStrict):
            name = palp.Field()
"""
from loguru import logger
from palp.item.item import ItemBase, Field


class StrictItemBase(ItemBase):
    """
        严格的 item，对 init 字段进行判断是否在运行的列表内
    """

    def __new__(cls, *args, **kwargs):
        """
        不使用 scrapy 的元类方式，但思想一致
        给每一个创建的类添加 palp_fields 字段来对输入字段去重

        :param args:
        :param kwargs:
        """
        cls.palp_fields = []

        for field, field_class in cls.__dict__.items():
            if isinstance(field_class, Field):
                cls.palp_fields.append(field)

        return object.__new__(cls)

    def __init__(self, **kwargs):
        for field, field_value in kwargs.items():
            if field not in self.__class__.palp_fields:
                logger.warning(f"字段 {field} 不在 {self.__class__.__name__} 允许字段列表中")
                continue

            self[field] = field_value


class StrictItem(StrictItemBase):
    """
        对 init 字段进行判断的 item
    """



"""
    spider 中间件
"""
from palp.spider.spider import Spider


class SpiderBaseMiddleware:
    """
        spider 中间件基类
    """

    def spider_start(self, spider: Spider) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """

    def spider_error(self, spider: Spider, exception: Exception) -> None:
        """
        spider 出错时的操作

        :param spider:
        :param exception: 错误
        :return:
        """

    def spider_close(self, spider: Spider) -> None:
        """
        spider 结束的操作

        :param spider:
        :return:
        """


class SpiderMiddleware(SpiderBaseMiddleware):
    """
        外部引用：请求中间件
    """

    def spider_start(self, spider) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """

    def spider_error(self, spider, exception: Exception) -> None:
        """
        spider 出错时的操作

        :param spider:
        :param exception: 错误
        :return:
        """

    def spider_close(self, spider) -> None:
        """
        spider 结束的操作

        :param spider:
        :return:
        """



"""
    单机 spider（不支持分布式）
"""
from palp import settings
from abc import abstractmethod
from palp.spider.spider import Spider
from palp.decorator.decorator_spider_middleware import SpiderMiddlewareDecorator
from palp.sequence.sequence_memory import PriorityMemorySequence, FIFOMemorySequence


class LocalSpider(Spider):
    """
        单机 spider（不支持分布式）
    """

    def __init__(self, thread_count: int = None, request_filter: bool = False, item_filter: bool = False):
        super().__init__(thread_count, request_filter, item_filter)
        setattr(settings, 'SPIDER_TYPE', 1)

        self.queue = PriorityMemorySequence()  # 请求队列
        self.queue_item = FIFOMemorySequence()  # item 队列

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """
        pass

    @SpiderMiddlewareDecorator()
    def run(self) -> None:
        self.start_controller()  # 任务处理
        self.start_distribute()  # 分发任务
        self.wait_distribute_thread_done()  # 等待任务分发完毕



"""
    Middleware

    注意：默认报错是 logger.error 输出，需要详细信息请使用 logger.exception（输出会很多，但很细）
"""
import palp
from typing import Union
from palp import settings
from loguru import logger
from palp.network.request import Request


class SpiderMiddleware(palp.SpiderMiddleware):
    """
        Spider 中间件：处理 spider 开启关闭任务
    """

    def spider_start(self, spider) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """

    def spider_error(self, spider, exception: Exception) -> None:
        """
        spider 出错时的操作

        :param spider:
        :param exception: 错误的详细信息
        :return:
        """
        logger.error(exception)

    def spider_close(self, spider) -> None:
        """
        spider 结束的操作

        :param spider:
        :return:
        """


class RequestMiddleware(palp.RequestMiddleware):
    """
        Request 中间件：处理请求
    """

    def request_in(self, spider, request) -> None:
        """
        请求进入时的操作

        :param spider:
        :param request:
        :return:
        """
        if settings.REQUEST_PROXIES_TUNNEL_URL:
            request.proxies = {
                'http': settings.REQUEST_PROXIES_TUNNEL_URL,
                'https': settings.REQUEST_PROXIES_TUNNEL_URL,
            }

    def request_error(self, spider, request, exception: Exception) -> Union[Request, None]:
        """
        请求出错时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param exception: 错误的详细信息
        :return: [Request, None]
        """
        logger.error(exception)

        return

    def request_failed(self, spider, request) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param request:
        :return:
        """
        logger.warning(f"失败的请求：{request}")

    def request_close(self, spider, request, response) -> Union[Request, None]:
        """
        请求结束时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param response:
        :return: [Request, None]
        """
        return



"""
    redis item 队列
    使用了 pickle 并使用 zlib 压缩
"""
from palp import settings
from palp.sequence.sequence_redis_request import FIFORequestRedisSequence, LIFORequestRedisSequence


class FIFOItemRedisSequence(FIFORequestRedisSequence):
    """
        先进先出队列
    """

    @classmethod
    def get_redis_key(cls):
        """
        获取 redis 的键

        :return:
        """
        return settings.REDIS_KEY_QUEUE_ITEM


class LIFOItemRedisSequence(LIFORequestRedisSequence, FIFOItemRedisSequence):
    """
        后进先出队列
    """



"""
    基础的 item

    要实现 Item() == {} 必须继承 MutableMapping 必须实现 item 的方法、__len__、__iter__ 方法

    注意：
        通过 isinstance 还是不能判断出是 dict，所以部分场景需要 to_dict

    案例：
        class Item(palp.Item):
            pass    # 此处啥都不用写
"""
import json
from typing import MutableMapping


class Field:
    """
        定义字段
    """


class ItemBase(MutableMapping):
    def to_dict(self) -> dict:
        """
        转化为 dict

        :return:
        """
        return dict(self)

    def to_json(self, **kwargs) -> str:
        """
        转化为 json

        :param kwargs: json 参数
        :return:
        """
        kwargs.setdefault('ensure_ascii', False)

        return json.dumps(self.to_dict(), **kwargs)

    def __setattr__(self, key, value):
        """
        设置属性（不提倡，但允许）

        :param key:
        :param value:
        :return:
        """
        self.__dict__[key] = value

    def __getattr__(self, item):
        """
        获取属性（不提倡，但允许）

        :param item:
        :return:
        """
        return self.__dict__[item]

    def __delattr__(self, item):
        """
        删除属性（不提倡，但允许）

        :param item:
        :return:
        """
        del self.__dict__[item]

    def __setitem__(self, key, value):
        """
        使类可以通过 xxx['xxx'] = xxx 进行设置

        :param key:
        :param value:
        :return:
        """
        self.__dict__[key] = value

    def __getitem__(self, item):
        """
        使类可以通过 xxx['xxx'] 进行访问

        :param item:
        :return:
        """
        return self.__dict__[item]

    def __delitem__(self, key):
        """
        使类可以通过 del xxx['xxx'] 进行移除

        :param key:
        :return:
        """
        del self.__dict__[key]

    def __len__(self):
        """
        使类可使用 len()

        :return:
        """
        return len(self.__dict__)

    def __iter__(self):
        """
        使类可遍历

        :return:
        """
        return iter(self.__dict__)

    def __getstate__(self):
        """
        pickle dumps 时使用

        :return:
        """
        return self.to_dict()

    def __setstate__(self, state):
        """
        pickle loads 时使用

        :param state:
        :return:
        """
        self.__dict__.update(state)

    def __str__(self):
        return f"<{self.__class__.__name__} item:{self.to_dict()}>"


class Item(ItemBase):
    """
        通用 item
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            self[key] = value
